/*
 * Module for looking up tags in MicroEmacs.
 * Earle R. Horton Sat, Oct 15, 1988.
 */
#include	<stdio.h>
#include	"estruct.h"
#include	"edef.h"
#define LSIZE	256	/* max. size of a line in the tags file */
#if LSC
#include	<resourcemgr.h>
#include	<strings.h>
#define c2pstr CtoPstr
#define p2cstr PtoCstr
#ifndef nil
#define nil 0L
#endif
#endif
#if MPW
#include	<resources.h>
#include	<string.h>
#endif
/*
 * tag(f,n) - command line version of dotag
 */
tag(f,n)
{
	dotag(TRUE,0);
}
/*
 * dotag(f,n) - goto tag
 * Should look in "tags," and then in the environment variable
 * "TAGS."  This looks in the MicroEMACS environment variable
 * "tags."  The Macintosh version looks for the "tags" file
 * in the same folder as the current buffer, which can be
 * confusing if the user is working with files in different
 * buffers.
 */
int
dotag(f,n)
{
	char	tag[NSTRING];
	int	status;
	char	*p;
	short	column;
	char c;
#if FINDER
/* Look first in folder with current file */
	extern short _iovrefnum;
	_iovrefnum = curbp->b_ioWDDirID;
#endif
	if(f){						/* Interactive. */
		if((status = mlreply("Tag: ",tag,NSTRING)) != TRUE) {
			mlwrite("[Aborted]");
			return(FALSE);
		}
	}else{				  /* Next identifier in buffer. */
		column = curwp->w_doto;
		c = lgetc(curwp->w_dotp,column);
		while(!isletter(c) && c != '_'){
			if(++column >= llength(curwp->w_dotp))
			return(FALSE);
			else c = lgetc(curwp->w_dotp,column);
		}
		p = tag;
		while(isletter(c) || c == '_'){
			*p++ = c;
			c = lgetc(curwp->w_dotp,++column);
			if(column > llength(curwp->w_dotp))
				break;
		}
		*p = '\0';
	}
	if(findtag(tag,"tags") == FALSE){
		if (findtag(tag,tagsfile) == FALSE){
			mlwrite("tag not found");
			status = FALSE;
		}
	}
	return status;
}
/*
 * findtag(tag,tfname) - does work for dotag()
 */
findtag(tag,tfname)
char	*tag;
char *tfname;
{
	char	lbuf[LSIZE],holdit[LSIZE];
	char	*fname, *str,*p;
	char	c;
	int		n;
	if (ffropen(tfname) != FIOSUC) {
		return FALSE;	/* FAILURE */
	}
	while (ffgetline(&n) == FIOSUC) {
		fline[n] = '\0';
		if ((fname = strchr(fline, '\t')) == NULL) {
			mlwrite("Format error in tags file");
			break;	/* FAILURE */
		}
		*fname++ = '\0';
		if ((str = strchr(fname, '\t')) == NULL) {
			mlwrite("Format error in tags file");
			break;	/* FAILURE */
		}
		*str++ = '\0';
		if (strcmp(fline, tag) == 0) {	/* Found the tag */
			strcpy(holdit,str);
			ffclose();
			strcpy(lbuf,"find-file ");
			strcat(lbuf,fname);
			execlevel = 0;
			docmd(lbuf);		/* open file with tag in it */
			gotobob(FALSE,0);	/* search from BOF */
			tagcmd(holdit,lbuf);	/* find the sucker */
			return TRUE;		/* SUCCESS */
		}
	}
	ffclose();
	return FALSE;	/* FAILURE */
}
/*
 * tagcmd(str,dst) - find the tag in the file
 */
tagcmd(src,dst)
char *src,*dst;
{
	char c;
	int haseol = FALSE;
	char *str = dst;
	char holdpat[NPAT];
	int holdmag;
	if(src[0] == '/' || src[0] == '?'){	/* Regular expression (func or macro) */
		++src;
		if(src[0] == '^'){
			*dst++ = EOLCHAR;
			src++;
		}
		while((c = *src++) != '\0'){
			switch(c){
				case '\\':
					*dst++ = *src++;
					break;
				case '$':	/* map '$' to EOLCHAR */
					if(src[0] == '/' || src[0] == '?'){
						*dst++ = EOLCHAR;
						haseol = TRUE;
					}
					else *dst++ = c;
					break;
				case '/':	/* ignore trailing separator */
				case '?':
					break;
				default:
					*dst++ = c;
					break;
			}
		}
		*dst++ = '\0';
		strcpy(holdpat,pat);		/* Save search pattern */
		holdmag = magical;		/* save magical */
		strcpy(pat,str);		/* put tag string in pat */
		magical = FALSE;		/* literal search */
		forwhunt(TRUE,1);		/* find it */
		if(haseol == TRUE) backline(TRUE,1);
		strcpy(pat,holdpat);		/* restore search variables */
		magical = holdmag;
	} else {				/* Typedef - line number */
		gotoline(TRUE,atoi(src));
	}
	gotobol(TRUE,1);
}
